import sys
from itertools import ifilter

from django.contrib.auth.signals import user_logged_in
import django.db.models

from django_otp.models import Device


DEVICE_ID_SESSION_KEY = 'otp_device_id'


def login(request, device):
    """
    Persist the given OTP device in the current session. The device will be
    rejected if it does not belong to ``request.user``.

    This is called automatically any time :func:`django.contrib.auth.login` is
    called with a user having an ``otp_device`` atribute.

    :param request: The HTTP request
    :type request: :class:`~django.http.HttpRequest`

    :param device: The OTP device used to authenticate the user.
    :type device: :class:`~django_otp.models.Device`
    """
    user = getattr(request, 'user', None)

    if (user is not None) and (device is not None) and (device.user_id == user.id):
        request.session[DEVICE_ID_SESSION_KEY] = device.persistent_id
        request.user.otp_device = device


def _handle_auth_login(sender, request, user, **kwargs):
    """
    Automatically persists an OTP device that was set by an OTP-aware
    AuthenticationForm.
    """
    if hasattr(user, 'otp_device'):
        login(request, user.otp_device)

user_logged_in.connect(_handle_auth_login)


def match_token(user, token):
    """
    Attempts to verify a :term:`token` on every device attached to the given
    user until one of them succeeds. As much as possible, you should prefer to
    verify tokens against specific devices.

    :param user: The user supplying the token.
    :type user: :class:`~django.contrib.auth.models.User`

    :param string token: An OTP token to verify.

    :returns: The device that accepted ``token``, if any.
    :rtype: :class:`~django_otp.models.Device` or ``None``
    """
    matches = ifilter(lambda d: d.verify_token(token), devices_for_user(user))

    return next(matches, None)


def devices_for_user(user, confirmed=True):
    """
    Returns an iterable of all devices registered to the given user.

    :param user:
    :type user: :class:`~django.contrib.auth.models.User`

    :param confirmed: If ``None``, all matching devices are returned.
        Otherwise, this can be any true or false value to limit the query
        to confirmed or unconfirmed devices, respectively.

    :rtype: iterable
    """
    for model_cls in device_classes():
        for device in model_cls.objects.devices_for_user(user, confirmed=confirmed):
            yield device


def device_classes():
    """
    Returns an iterable of all loaded device models.
    """
    return ifilter(lambda m: issubclass(m, Device), django.db.models.get_models())


def import_class(path):
    """
    Imports a class based on a full Python path ('pkg.pkg.mod.Class'). This
    does not trap any exceptions if the path is not valid.
    """
    module, name = path.rsplit('.', 1)
    __import__(module)
    mod = sys.modules[module]
    cls = getattr(mod, name)

    return cls



from django.test import TestCase
from django.core import mail

from django_otp.forms import OTPAuthenticationForm
from django_otp.plugins.otp_static.models import StaticDevice
from django_otp.plugins.otp_email.models import EmailDevice


class AuthFormTest(TestCase):
    fixtures = ['django_otp/alice_and_bob.yaml']

    def test_empty(self):
        data = {}
        form = OTPAuthenticationForm(None, data)

        self.assert_(not form.is_valid())
        self.assertEqual(form.get_user(), None)

    def test_bad_password(self):
        data = {
            'username': 'alice',
            'password': 'bogus',
        }
        form = OTPAuthenticationForm(None, data)

        self.assert_(not form.is_valid())
        self.assert_(form.get_user() is None)
        self.assertEqual(form.errors.keys(), ['__all__'])

    def test_no_token(self):
        data = {
            'username': 'alice',
            'password': 'password',
        }
        form = OTPAuthenticationForm(None, data)

        self.assert_(not form.is_valid())
        self.assert_(form.get_user().username == 'alice')

    def test_passive_token(self):
        data = {
            'username': 'alice',
            'password': 'password',
            'otp_token': 'alice1',
        }
        form = OTPAuthenticationForm(None, data)

        self.assert_(form.is_valid())
        alice = form.get_user()
        self.assert_(alice.username == 'alice')
        self.assert_(isinstance(alice.otp_device, StaticDevice))
        self.assertEqual(alice.otp_device.token_set.count(), 2)

    def test_spoofed_device(self):
        data = {
            'username': 'alice',
            'password': 'password',
            'otp_device': 'django_otp.plugins.otp_static.models.StaticDevice/10',
            'otp_token': 'bob1',
        }
        form = OTPAuthenticationForm(None, data)

        self.assert_(not form.is_valid())
        alice = form.get_user()
        self.assert_(alice.username == 'alice')
        self.assert_(alice.otp_device is None)

    def test_specific_device_fail(self):
        data = {
            'username': 'alice',
            'password': 'password',
            'otp_device': 'django_otp.plugins.otp_email.models.EmailDevice/1',
            'otp_token': 'alice1',
        }
        form = OTPAuthenticationForm(None, data)

        self.assert_(not form.is_valid())
        alice = form.get_user()
        self.assert_(alice.username == 'alice')
        self.assert_(alice.otp_device is None)

    def test_specific_device(self):
        data = {
            'username': 'alice',
            'password': 'password',
            'otp_device': 'django_otp.plugins.otp_static.models.StaticDevice/1',
            'otp_token': 'alice1',
        }
        form = OTPAuthenticationForm(None, data)

        self.assert_(form.is_valid())
        alice = form.get_user()
        self.assert_(alice.username == 'alice')
        self.assert_(alice.otp_device is not None)

    def test_email_interaction(self):
        data = {
            'username': 'alice',
            'password': 'password',
            'otp_device': 'django_otp.plugins.otp_email.models.EmailDevice/1',
            'otp_token': '',
            'otp_challenge': '1',
        }
        form = OTPAuthenticationForm(None, data)

        self.assert_(not form.is_valid())
        alice = form.get_user()
        self.assert_(alice.username == 'alice')
        self.assert_(alice.otp_device is None)
        self.assertEqual(len(mail.outbox), 1)

        data['otp_token'] = mail.outbox[0].body
        del data['otp_challenge']
        form = OTPAuthenticationForm(None, data)

        self.assert_(form.is_valid())
        self.assert_(isinstance(form.get_user().otp_device, EmailDevice))



from django.utils import unittest
from doctest import DocTestSuite

from django_otp import util
from django_otp import oath

from .forms import AuthFormTest


def suite():
    loader = unittest.TestLoader()
    suite = unittest.TestSuite()

    suite.addTest(DocTestSuite(util))
    suite.addTest(DocTestSuite(oath))
    suite.addTest(loader.loadTestsFromTestCase(AuthFormTest))

    return suite



from django.test import TestCase

from .models import HOTPDevice


class HOTPTest(TestCase):
    fixtures = ['tests/hotp']

    # The next three tokens
    tokens = [782373, 313268, 307722]

    def setUp(self):
        self.device = HOTPDevice.objects.get()

    def test_normal(self):
        ok = self.device.verify_token(self.tokens[0])

        self.assert_(ok)
        self.assertEqual(self.device.counter, 1)

    def test_normal_drift(self):
        ok = self.device.verify_token(self.tokens[1])

        self.assert_(ok)
        self.assertEqual(self.device.counter, 2)

    def test_excessive_drift(self):
        ok = self.device.verify_token(self.tokens[2])

        self.assert_(not ok)
        self.assertEqual(self.device.counter, 0)

    def test_bad_value(self):
        ok = self.device.verify_token(123456)

        self.assert_(not ok)
        self.assertEqual(self.device.counter, 0)



from time import time

from django.test import TestCase

from .models import TOTPDevice
from .conf import settings


class TOTPTest(TestCase):
    fixtures = ['tests/totp']

    # The next ten tokens
    tokens = [179225, 656163, 839400, 154567, 346912, 471576, 45675, 101397, 491039, 784503]

    def setUp(self):
        """
        Load the device and move it to the fourth time step. The current token
        is 154567.
        """
        settings.OTP_TOTP_SYNC = False
        self.device = TOTPDevice.objects.get()
        self.device.t0 = int(time() - (30 * 3))

    def test_single(self):
        results = [self.device.verify_token(token) for token in self.tokens]

        self.assertEqual(results, [False]*3 + [True] + [False]*6)

    def test_tolerance(self):
        self.device.tolerance = 1
        results = [self.device.verify_token(token) for token in self.tokens]

        self.assertEqual(results, [False]*2 + [True]*3 + [False]*5)

    def test_drift(self):
        self.device.tolerance = 1
        self.device.drift = -1
        results = [self.device.verify_token(token) for token in self.tokens]

        self.assertEqual(results, [False]*1 + [True]*3 + [False]*6)

    def test_sync_drift(self):
        settings.OTP_TOTP_SYNC = True
        self.device.tolerance = 2
        ok = self.device.verify_token(self.tokens[5])

        self.assert_(ok)
        self.assertEqual(self.device.drift, 2)

    def test_sync_results(self):
        self.device.tolerance = 1
        settings.OTP_TOTP_SYNC = True
        self.device.verify_token(self.tokens[4])
        settings.OTP_TOTP_SYNC = False
        results = [self.device.verify_token(token) for token in self.tokens]

        self.assertEqual(self.device.drift, 1)
        self.assertEqual(results, [False]*3 + [True]*3 + [False]*4)
